"""
Between-levels screen.
"""
import pygame
import Player
import Sprite
import Sound
import Alien
import time
from Constants import *
from Utils import *

AboutText = """Trouble Shooter (v20041017)
@
Trouble Shooter is Dumple's entry in the Ludum Dare 48-hour
game design competition.  It was developed in Python using
the Pygame development library.  Music written in ModPlug,
noises developed in GoldWave, and graphics
done with a mix of code and PhotoShop.
@
Gratuitous plug: Check out Microbattle (another 48-hour
contest game) and Tendrils (the rhythmic combat RPG).
@
@
--Dumple

"""

InstructionsText = """Welcome to Trouble Shooter.
@
You are a Space Viking, traveling through dimensions in search
of treasure and adventure.  No two sectors are quite alike, and
even the laws of physics may change!  The SPACE POPE governs
this galaxy, and you must obey his edicts, at least for now...
@
Turn your ship and activate thrusters with the arrow keys.  Use
left CTRL to fire weapons.  Press ALT to either pick up passengers,
or strafe, or fire emergency missiles...depending on the level.
@
Some monsters may drop SPACE CAPSULES.  Some capsules are good,
some bad.  Check the color carefully - but be careful, because
the color code differs from one game to the next!
@
Each sector has an OBJECTIVE (shown during the game at the bottom
of the screen).  If it's a KILL mission, shoot that kind of
alien (and anyone else who gets in your way).  If it's a RESCUE
mission, don't shoot that color of alien - fly on top of them and
press ALT.  And finally, on a black hole mission, pick up the alien,
then press ALT again to toss them into the black hole.  (Avoid flying
over the black hole yourself, it drains energy!)
"""

class StarSprite(Sprite.AnimateSprite):
    def __init__(self):
        self.DX = random.randrange(-50, 51) / 10.0
        self.DY = random.randrange(-50, 51) / 10.0
        self.X = self.DX*5 + Global.App.Width / 2
        self.Y = self.DY*5 + Global.App.Height / 2
        Sprite.AnimateSprite.__init__(self, ImageNames.Star, self.X, self.Y)
    def Update(self):
        self.X += self.DX
        self.Y += self.DY
        if self.X < 0 or self.X > 1024:
            self.kill()
            return
        if self.Y < 0 or self.Y > 768:
            self.kill()
            return
        self.rect.center = (self.X, self.Y)
        Sprite.AnimateSprite.Update(self)

class IntroStates:
    SpacePope = 0 # Spaaaaaaaaaace poooooope!
    Instructions = 1
    About = 2

class LevelIntroScreen:
    def __init__(self, App):
        self.Surface = App.Surface
        self.BackgroundSurface = App.BackgroundSurface
        self.BackgroundSprites = pygame.sprite.RenderUpdates()
        self.ForegroundSprites = pygame.sprite.RenderUpdates()
        self.AnimationSprites = pygame.sprite.RenderUpdates()
        self.StarSprites = pygame.sprite.RenderUpdates()
        self.ButtonSprites = pygame.sprite.RenderUpdates()
        self.Tick = 0
        self.ScreenClosed = 0
        self.MaxTick = 2048
        self.SongTimer = 20
        self.State = IntroStates.SpacePope
        self.PopeSprite = None
    def RedrawSelf(self):
        self.BackgroundSurface.fill(Color.Black)
        #self.BackgroundSurface.blit(self.BackingSprite.image, (0,0))
        for Sprite in self.BackgroundSprites.sprites():
            self.BackgroundSprites.draw(self.BackgroundSurface)
        MarkDirty([(0, 0,Global.App.Width,Global.App.Height)])
        self.Surface.blit(self.BackgroundSurface,(0,0))
        
    def HandleEvents(self):
        for Event in pygame.event.get():
            self.HandleEvent(Event)
    def HandleEvent(self, Event):
        if Event.type is pygame.QUIT:
            Global.App.QuittingTime = 1
        elif Event.type == EndSongEvent:
            Sound.RepeatSong()
        elif Event.type is pygame.KEYDOWN:
            self.HandleKey(Event.key)
        elif Event.type is pygame.MOUSEBUTTONDOWN:
            self.HandleMouseClicked(Event.pos, Event.button)
            return
    def HandleMouseClicked(self, Pos, Button):
        Sprite = CollidePointWithGroup(Pos, self.ButtonSprites)
        if Sprite:
            if Sprite.Text == "Done":
                self.State = IntroStates.SpacePope
                self.Render()
            elif Sprite.Text == "Instructions":
                self.State = IntroStates.Instructions
                self.Render()
            elif Sprite.Text == "About":
                self.State = IntroStates.About
                self.Render()
            elif Sprite.Text == "Start!":
                self.DestroyScreen()
            elif Sprite.Text == "Quit!":
                Global.App.QuittingTime = 1
            elif Sprite.Text == "Next Sector":
                Global.Player.SectorNumber = min(Global.Player.SectorNumber + 1, Global.Player.MaxSectorNumber)
                Global.App.GetObjective(Global.Player.SectorNumber)
                self.Render()
            elif Sprite.Text == "Prev Sector":
                Global.Player.SectorNumber = max(Global.Player.SectorNumber - 1, 1)
                Global.App.GetObjective(Global.Player.SectorNumber)
                self.Render()
                
    def HandleKey(self, Key):
        if Key == 13:
            KeyMods = pygame.key.get_mods()
            if (KeyMods & (pygame.KMOD_RALT | pygame.KMOD_LALT)):
                Global.App.ToggleFullScreen()
        elif Key is pygame.K_ESCAPE:
            if self.State == IntroStates.SpacePope:
                Global.App.QuittingTime = 1
            else:
                self.State = IntroStates.SpacePope
                self.Render()
        elif Key == ord(" "):
            self.DestroyScreen()
        elif Key == ord("i") and self.State == IntroStates.SpacePope:
            self.State = IntroState.Instructions
            self.Render()
        elif Key == ord("a") and self.State == IntroStates.SpacePope:
            self.State = IntroState.About
            self.Render()
            
    def DestroyScreen(self):
        for Sprite in self.ForegroundSprites.sprites():
            Sprite.kill()
        for Sprite in self.BackgroundSprites.sprites():
            Sprite.kill()
        self.ScreenClosed = 1
            
    def Render(self):
        if self.PopeSprite:
            self.PopeSprite.kill()
            self.PopeSprite = None
        for Sprite in self.BackgroundSprites.sprites():
            Sprite.kill() # Clean!
        if self.State == IntroStates.SpacePope:
            self.RenderSpacePope()
        elif self.State == IntroStates.Instructions:
            self.RenderInstructions()
        else:
            self.RenderAbout()
        self.RedrawSelf()
    def RenderTextBlok(self, Text):
        Words = Text.split()
        Y = 60
        CurrentStr = ""
        for Word in Words:
            if Word == "@":
                if CurrentStr:
                    TextSprite = Sprite.TextSprite(CurrentStr, 512, Y, FontSize = 24)
                    self.BackgroundSprites.add(TextSprite)
                    Y += 20
                    CurrentStr = ""
                Y += 20
                continue
            if len(CurrentStr + Word)> 80:
                TextSprite = Sprite.TextSprite(CurrentStr, 512, Y, FontSize = 24)
                self.BackgroundSprites.add(TextSprite)
                Y += 20
                CurrentStr = ""                
            CurrentStr += " "+Word
        if CurrentStr:
            TextSprite = Sprite.TextSprite(CurrentStr, 512, Y, FontSize = 24)
            self.BackgroundSprites.add(TextSprite)
            Y += 20
    def RenderAbout(self):
        self.RenderTextBlok(AboutText)
        # BUTTONS:
        ButtonSprite = Sprite.ButtonSprite("Done", 512, 650)
        self.BackgroundSprites.add(ButtonSprite)
        self.ButtonSprites.add(ButtonSprite)
    def RenderInstructions(self):
        self.RenderTextBlok(InstructionsText)
        # BUTTONS:
        ButtonSprite = Sprite.ButtonSprite("Done", 512, 650)
        self.BackgroundSprites.add(ButtonSprite)
        self.ButtonSprites.add(ButtonSprite)
    def RenderSpacePope(self):
        random.seed(Global.Player.SectorNumber)
        TSprite = Sprite.TextSprite("Sector %d"%Global.Player.SectorNumber, 512, 50, FontSize = 24)
        self.BackgroundSprites.add(TSprite)
        if random.random() > 0.5:
            FirstWords = ("Final", "Dark", "Forgotten", "Ninja", "Deadly", "Merciless")
            LastWords = ("Justice", "Karma", "Fate", "Dragon", "Loaf", "Destiny")
            SectorTitle = "%s %s"%(random.choice(FirstWords), random.choice(LastWords))
        else:
            ThingyWords = ("Nebula", "Dimension", "Vortex", "Realm")
            Adjectives = ("", "", "", "", "",
                          "Plaid ", "Space ", "Mutant ", "Sexy ", "Lazy ",
                          "Angry ", "Purple ", "Glowing ")
            Objects = ("Lightning", "Camels", "Critters", "Zombies", "Bees",
                       "Tyrants", "Sorcery")
            SectorTitle = "The %s of %s%s"%(random.choice(ThingyWords),
                                            random.choice(Adjectives),
                                            random.choice(Objects))
        TSprite = Sprite.TextSprite(SectorTitle, 512, 100, FontSize = 32)
        self.BackgroundSprites.add(TSprite)
        TSprite = Sprite.TextSprite("Your Mission:", 250, 300, FontSize = 32)
        self.BackgroundSprites.add(TSprite)
        ObjectiveText = GetObjectiveDescription(Global.Objective)
        TSprite = Sprite.TextSprite(ObjectiveText, 250, 350, FontSize = 32)
        self.BackgroundSprites.add(TSprite)
        self.PopeSprite = Sprite.AnimateSprite("SpacePope", 800, 300)
        self.ForegroundSprites.add(self.PopeSprite)
        TSprite = Sprite.TextSprite("Space Pope says:", 800, 400, FontSize = 32)
        self.BackgroundSprites.add(TSprite)
        TSprite = Sprite.TextSprite(Global.Rules.Gravity, 800, 440, Color.Red, FontSize = 32)
        self.BackgroundSprites.add(TSprite)
        TSprite = Sprite.TextSprite("Button2 is %s"%Global.Rules.Button2, 800, 480, Color.Red, FontSize = 32)
        self.BackgroundSprites.add(TSprite)
        TSprite = Sprite.TextSprite("Down is %s"%Global.Rules.Down, 800, 520, Color.Red, FontSize = 32)
        self.BackgroundSprites.add(TSprite)
        ZanyRule = random.choice(("","","","",
                                  "Don't shoot food", "Watch out for snakes"))
        TSprite = Sprite.TextSprite(ZanyRule, 800, 560, Color.Red, FontSize = 32)
        self.BackgroundSprites.add(TSprite)
        TSprite = Sprite.TextSprite("Press [Space] to start level", 512, 700, FontSize = 32)
        self.BackgroundSprites.add(TSprite)
        # BUTTONS:
        ButtonSprite = Sprite.ButtonSprite("Instructions", 312, 600)
        self.BackgroundSprites.add(ButtonSprite)
        self.ButtonSprites.add(ButtonSprite)
        ButtonSprite = Sprite.ButtonSprite("About", 512, 600)
        self.BackgroundSprites.add(ButtonSprite)
        self.ButtonSprites.add(ButtonSprite)
        ButtonSprite = Sprite.ButtonSprite("Quit!", 712, 600)
        self.BackgroundSprites.add(ButtonSprite)
        self.ButtonSprites.add(ButtonSprite)
        if Global.Player.SectorNumber < Global.Player.MaxSectorNumber:
            ButtonSprite = Sprite.ButtonSprite("Next Sector", 600, 660)
            self.BackgroundSprites.add(ButtonSprite)
            self.ButtonSprites.add(ButtonSprite)
        if Global.Player.SectorNumber > 1:
            ButtonSprite = Sprite.ButtonSprite("Prev Sector", 400, 660)
            self.BackgroundSprites.add(ButtonSprite)
            self.ButtonSprites.add(ButtonSprite)
            
        self.RedrawSelf()
    def Update(self):
        if self.SongTimer:
            self.SongTimer -= 1
            if self.SongTimer == 0:
                Sound.PlaySong("Title.xm")
        # This screen is boring.  But we do have stars.
        if len(self.StarSprites.sprites()) < 50:
            Sprite = StarSprite()
            self.StarSprites.add(Sprite)
            self.AnimationSprites.add(Sprite)
        for Sprite in self.ForegroundSprites.sprites():
            Sprite.Update()
        for Sprite in self.AnimationSprites.sprites():
            Sprite.Update()
            
    def HandleLoop(self):
        self.Tick = (self.Tick + 1) % self.MaxTick
        self.AnimationSprites.clear(self.Surface, self.BackgroundSurface)
        self.ForegroundSprites.clear(self.Surface, self.BackgroundSurface)
        self.Update()
        if self.ScreenClosed:
            return
        self.HandleEvents()
        if self.ScreenClosed:
            return
        DirtRects = self.ForegroundSprites.draw(self.Surface)
        MarkDirty(DirtRects)
        DirtRects = self.AnimationSprites.draw(self.Surface)
        MarkDirty(DirtRects)

        