from Utils import *
from Constants import *
import Sound

import Sprite
import Global
import math

XDir =  [0.00, -0.20, -0.38, -0.56, -0.71, -0.83, -0.92, -0.98,
    -1.00, -0.98, -0.92, -0.83, -0.71, -0.56, -0.38, -0.20, -0.00,
    0.20, 0.38, 0.56, 0.71, 0.83, 0.92, 0.98, 1.00, 0.98, 0.92,
    0.83, 0.71, 0.56, 0.38, 0.20]

YDir = [-1.00, -0.98, -0.92, -0.83, -0.71, -0.56, -0.38, -0.20, 0.00, 0.20,
        0.38, 0.56, 0.71, 0.83, 0.92, 0.98, 1.00, 0.98, 0.92, 0.83, 0.71, 0.56,
        0.38, 0.20, 0.00, -0.20, -0.38, -0.56, -0.71, -0.83, -0.92, -0.98, ]
DirCount = 32

class ShieldSprite(Sprite.AnimateSprite):
    def __init__(self):
        Sprite.AnimateSprite.__init__(self, ImageNames.Shield, Global.PlayerSprite.X, Global.PlayerSprite.Y)

class PlayerSprite(Sprite.Sprite, Sprite.Bouncer):
    BaseMaximumSpeed = 4
    def __init__(self, X, Y):
        Sprite.Sprite.__init__(self, "Ship", X, Y)
        self.ShieldSprite = None
        self.MaximumSpeed = self.BaseMaximumSpeed
        self.Images = Global.Images[ImageNames.Ship]
        self.Dir = 0 # Direction we're pointing
        self.XDir = XDir[self.Dir] # our acceleration-dir
        self.YDir = YDir[self.Dir] # our acceleration-dir
        self.DX = 0 # direction we're MOVING.
        self.DY = 0
        self.X = X
        self.Y = Y
        self.Invincible = 0
        self.CarriedCritter = None
        self.Energy = 1000
    def RotateRight(self):
        if Global.Screen.StrafeAvailable and pygame.key.get_pressed()[308]:
            StrafeDir = (self.Dir - 8)
            if StrafeDir<0:
                StrafeDir += DirCount            
            self.DX += XDir[StrafeDir] * 0.2
            self.DY += YDir[StrafeDir] * 0.2
            Speed = math.sqrt(self.DX*self.DX + self.DY*self.DY)
            if Speed > self.MaximumSpeed:
                self.DX = self.DX * (self.MaximumSpeed / Speed)
                self.DY = self.DY * (self.MaximumSpeed / Speed)
            TSprite = Sprite.ThrustSprite(self.X - XDir[StrafeDir]*6, self.Y - YDir[StrafeDir]*6,
                                          -XDir[StrafeDir], -YDir[StrafeDir])
            Global.Screen.AnimationSprites.add(TSprite)            
        else:
            if self.Dir == 0:
                self.Dir = (DirCount - 1)
            else:
                self.Dir = (self.Dir - 1)
            self.ReplaceImage(self.Images[self.Dir])
            self.XDir = XDir[self.Dir]
            self.YDir = YDir[self.Dir]
    def RotateLeft(self):
        if Global.Screen.StrafeAvailable and pygame.key.get_pressed()[308]:
            StrafeDir = (self.Dir + 8) % DirCount            
            self.DX += XDir[StrafeDir] * 0.2
            self.DY += YDir[StrafeDir] * 0.2
            Speed = math.sqrt(self.DX*self.DX + self.DY*self.DY)
            if Speed > self.MaximumSpeed:
                self.DX = self.DX * (self.MaximumSpeed / Speed)
                self.DY = self.DY * (self.MaximumSpeed / Speed)
            TSprite = Sprite.ThrustSprite(self.X - XDir[StrafeDir]*6, self.Y - YDir[StrafeDir]*6,
                                          -XDir[StrafeDir], -YDir[StrafeDir])
            Global.Screen.AnimationSprites.add(TSprite)            
        else:        
            self.Dir = (self.Dir + 1)%DirCount
            self.ReplaceImage(self.Images[self.Dir])
            self.XDir = XDir[self.Dir]
            self.YDir = YDir[self.Dir]
    def UpdateSpeedLimit(self):
        self.MaximumSpeed = self.BaseMaximumSpeed 
        if Global.Player.Conditions.get(Conditions.Haste):
            self.MaximumSpeed *= 1.5
        if Global.Player.Conditions.get(Conditions.Slow):
            self.MaximumSpeed *= 0.5
    def Thrust(self):
        self.DX += self.XDir * 0.2
        self.DY += self.YDir * 0.2
        Speed = math.sqrt(self.DX*self.DX + self.DY*self.DY)
        if Speed > self.MaximumSpeed:
            self.DX = self.DX * (self.MaximumSpeed / Speed)
            self.DY = self.DY * (self.MaximumSpeed / Speed)
        TSprite = Sprite.ThrustSprite(self.X - self.XDir*6, self.Y - self.YDir*6, -self.XDir, -self.YDir)
        Global.Screen.AnimationSprites.add(TSprite)
    def Brake(self):
        self.DX -= self.XDir * 0.2
        self.DY -= self.YDir * 0.2
        Speed = math.sqrt(self.DX*self.DX + self.DY*self.DY)
        if Speed > self.MaximumSpeed:
            self.DX = self.DX * (self.MaximumSpeed / Speed)
            self.DY = self.DY * (self.MaximumSpeed / Speed)
        if random.random() > 0.7:
            TSprite = Sprite.ThrustSprite(self.X + self.XDir*6, self.Y + self.YDir*6, +self.XDir, +self.YDir)
            Global.Screen.AnimationSprites.add(TSprite)
            
    def HandleShot(self, Shot):
        if self.ShieldSprite:
            Global.Player.Energy -= Shot.Damage / 10
        else:
            Global.Player.Energy -= Shot.Damage
        Sound.PlaySound(SoundNames.PlayerHit)
        self.DX += Shot.DX
        self.DY += Shot.DY
        Global.Screen.CheckDeath()
    def SetImage(self):
        self.ReplaceImage(self.Images[self.Dir])
    def InvisoImage(self):
        self.ReplaceImage(Global.NullImage)
    def Update(self):
        self.Invincible = max(0, self.Invincible - 1)
        self.X += self.DX
        self.Y += self.DY
        self.rect.center = (self.X, self.Y)
        if self.ShieldSprite:
            self.ShieldSprite.rect.center = self.rect.center
        
##            if Global.Player.HP < Global.Player.Caution:
##                self.Meep -= Global.TickSpeed
##                if self.Meep <= 0:
##                    Sound.PlaySound("meep")
##                    self.Meep = self.MaxMeep
        