"""
Trouble Shooter [tm] is a game of interstellar cunning and intrigue, and
also sometimes stuff blows up for no reason.

The main idea: Game is a series of levels which tend to mostly increase
in difficulty.  The gameplay is asteroids-y (you are a "ship" that thrusts
around the screen), but the controls vary a bit from one level to the next.
The laws of physics change a bit between levels.  And, the objectives change
between levels.  (Ideally, it should be a 2-player game, so that each
player faces the same set of challenges...)

Types of sprites:
WallSprites
HazardSprites (black holes, planets, chests - things that don't move but can interact w/player)
PlayerSprite (could be 2 of these for 2player mode)
PlayerShotSprites and MonsterShotSprites (some sprites live in both groups)
BouncySprites
--MonsterSprites
--ItemSprites (these bounce around)
"""
import random # YAY!  I FIT THE THEME!  YOU GIVE ME VOTES NOW!
import pygame
import Player
import Sprite
import Sound
import Alien
import time
import Items
import Snake
import traceback
from Constants import *
from Utils import *
class Activity:
    Play = 0
    Pause = 1
    DeathDelay = 2
    LevelComplete = 3
    StartLevelDelay = 4

class SpaceScreen:
    MaxTick = 2048
    Bottom = 700
    def __init__(self, App):
        self.Surface = App.Surface
        self.BackgroundSurface = App.BackgroundSurface
        self.SnakeSprites = pygame.sprite.RenderUpdates()
        self.BackgroundSprites = pygame.sprite.RenderUpdates()
        self.ForegroundSprites = pygame.sprite.RenderUpdates()
        self.AnimationSprites = pygame.sprite.RenderUpdates()
        self.MonsterSprites = pygame.sprite.Group()
        self.AllShotSprites = pygame.sprite.Group()
        self.MonsterShotSprites = pygame.sprite.Group()
        self.PlayerShotSprites = pygame.sprite.Group()
        self.PanicSprites = pygame.sprite.Group()
        self.BouncySprites = pygame.sprite.Group()
        self.WallSprites = pygame.sprite.Group()
        self.ItemSprites = pygame.sprite.Group()
        self.StatusPanelSprites = pygame.sprite.Group()
        self.HazardSprites = pygame.sprite.Group()
        self.ShrapnelSprites = pygame.sprite.Group()
        self.StatusPanelSprite = None
        self.EnergySprite = None
        self.ScoreSprite = None
        self.Tick = 0
        self.ShieldsAvailable = 0
        self.StrafeAvailable = 0
        self.BackingSprite = None
        self.Activity = Activity.Play
        self.ScreenClosed = 0
        self.Score = 0
        self.LifeCount = 3
        self.Energy = 1000
        self.KeyHandlers = {}
        self.KeyDelays = {} # you see, billy, even atomic starships have a refractory period...
        self.ShotDelay = 0
        self.KeyRefractory = {}
        self.DeathDelayCounter = 0
        self.Countdown = 0
        self.ItemRate = 0.03 # default
        self.ObjectiveProgressSprite = None
        for Command in KeyCommand.AllCommands:
            self.KeyDelays[Command] = 0
            self.KeyRefractory[Command] = 2 #default
        self.KeyRefractory[KeyCommand.Up] = 1
        self.KeyRefractory[KeyCommand.Shot1] = 30
    def RedrawSelf(self):
        self.BackgroundSurface.fill(Color.Black)
        if not self.BackingSprite:
            self.DrawBacking()
        self.BackgroundSurface.blit(self.BackingSprite.image, (0,0))
        for Sprite in self.BackgroundSprites.sprites():
            self.BackgroundSprites.draw(self.BackgroundSurface)
        MarkDirty([(0, 0,Global.App.Width,Global.App.Height)])
        self.Surface.blit(self.BackgroundSurface,(0,0))
    def DrawBacking(self):
        # Draw some kroonchy stars.
        if not self.BackingSprite:
            Image = pygame.Surface((Global.App.Width, self.Bottom))
            self.BackingSprite = Sprite.Sprite(Image, 0, 0)
        self.BackingSprite.image.fill(Color.Black)
        for StarIndex in range(300):
            X = random.randrange(Global.App.Width)
            Y = random.randrange(self.Bottom)
            Image = random.choice(Global.Images[ImageNames.Star])
            self.BackingSprite.image.blit(Image, (X, Y))
    def AddShrapnel(self, X, Y, Count):
        if len(self.ShrapnelSprites.sprites()) < 50:
            for Index in range(Count):
                SSprite = Sprite.ShrapnelSprite(X, Y)
                self.ShrapnelSprites.add(SSprite)
                self.AnimationSprites.add(SSprite)
    def SelectAlienBehavior(self, SectorNumber):
        AlienClasses = [Alien.RedAlienSprite, Alien.GreenAlienSprite, Alien.BlueAlienSprite]
        if Global.Objective.Class in AlienClasses:
            AlienClasses.remove(Global.Objective.Class)
        random.shuffle(AlienClasses)
        if Global.Objective.Class and (Global.Objective.Class not in AlienClasses):
            AlienClasses.append(Global.Objective.Class)
            AlienClasses.reverse()
        if Global.Objective.Type in (ObjectiveType.Rescue, ObjectiveType.BlackHoleToss):
            Behaviors = [0]
        else:
            # Choose anything but a splitter:
            Behaviors = [random.choice((1,2,3,4))]
        Others = random.choice([(1,2), (1,3), (1,4),
                               (3,4),
                               (5,1), (6,4), (6,5), (6,5), (6,5)])
        Behaviors.extend(Others)
        for Index in range(3):
            AlienClass = AlienClasses[Index]
            Behavior = Behaviors[Index]
            SpeedFactor = 20.0
            if Behavior == 0:
                # PEACEFUL
                AlienClass.SpeedLimit = random.randrange(10, 30) / SpeedFactor
                AlienClass.MaxHP = 1
                AlienClass.IsHostile = 0
                AlienClass.CanShoot = 0
                AlienClass.Update = Alien.AlienSprite.UpdatePeaceful
                AlienClass.HandleDeath = Alien.AlienSprite.DieStandard
                AlienClass.Init = Alien.AlienSprite.InitPeaceful
                AlienClass.SplitCount = 0
                AlienClass.SplitClass = None
                
            elif Behavior == 1:
                # HOSTILE
                AlienClass.SpeedLimit = random.randrange(10, 30) / SpeedFactor
                AlienClass.MaxHP = 1
                AlienClass.IsHostile = 1
                AlienClass.CanShoot = 0
                AlienClass.Init = Alien.AlienSprite.Init
                AlienClass.Update = Alien.AlienSprite.UpdateHostile
                AlienClass.HandleDeath = Alien.AlienSprite.DieStandard
                AlienClass.SplitCount = 0
                AlienClass.SplitClass = None
                
            elif Behavior == 2:
                # SHOOTER
                AlienClass.SpeedLimit = random.randrange(5, 20) / SpeedFactor
                AlienClass.MaxHP = 1
                AlienClass.IsHostile = 1
                AlienClass.CanShoot = 1
                AlienClass.Init = Alien.AlienSprite.InitShooter
                AlienClass.Update = Alien.AlienSprite.UpdateShooter
                AlienClass.HandleDeath = Alien.AlienSprite.DieStandard
                AlienClass.SplitCount = 0
                AlienClass.SplitClass = None
                
            elif Behavior == 3:
                # SNIPER
                AlienClass.SpeedLimit = random.randrange(10, 30) / SpeedFactor
                AlienClass.MaxHP = 1
                AlienClass.IsHostile = 1
                AlienClass.CanShoot = 0
                AlienClass.Init = Alien.AlienSprite.InitSniper
                AlienClass.Update = Alien.AlienSprite.UpdateSniper
                AlienClass.HandleDeath = Alien.AlienSprite.DieStandard
                AlienClass.SplitCount = 0
                AlienClass.SplitClass = None
                
            elif Behavior == 4:
                # SPEEDY
                AlienClass.SpeedLimit = random.randrange(30, 50) / SpeedFactor
                AlienClass.MaxHP = 1
                AlienClass.IsHostile = 1
                AlienClass.CanShoot = 0
                AlienClass.Init = Alien.AlienSprite.Init
                AlienClass.Update = Alien.AlienSprite.UpdateHostile
                AlienClass.HandleDeath = Alien.AlienSprite.DieStandard
                AlienClass.SplitCount = 0
                AlienClass.SplitClass = None
            elif Behavior == 5:
                # SPLITTER-2:
                AlienClass.SpeedLimit = random.randrange(10, 20) / SpeedFactor
                AlienClass.MaxHP = 1
                AlienClass.IsHostile = 1
                AlienClass.CanShoot = 0
                AlienClass.Init = Alien.AlienSprite.InitSniper
                AlienClass.Update = Alien.AlienSprite.UpdateSniper
                AlienClass.HandleDeath = Alien.AlienSprite.DieSplit
                AlienClass.SplitCount = 2
                AlienClass.SplitClass = AlienClasses[(Index+1)%3]
            elif Behavior == 6:
                # SPLITTER-5:
                AlienClass.SpeedLimit = random.randrange(5, 15) / SpeedFactor
                AlienClass.MaxHP = random.randrange(3, 7)
                AlienClass.IsHostile = 1
                AlienClass.CanShoot = 0
                AlienClass.Init = Alien.AlienSprite.InitBouncy
                AlienClass.Update = Alien.AlienSprite.UpdateBounce
                AlienClass.HandleDeath = Alien.AlienSprite.DieSplit
                AlienClass.SplitCount = 5
                AlienClass.SplitClass = AlienClasses[(Index+1)%3]
        
    def BuildSpawners(self):
        Y = 175
        # There are nine loci where we can stick things.  Could be a black hole,
        # or a spawner, or a wall, or nothing at all.
        Loci = [(256, Y), (512, Y), (778, Y),
                (256, Y*2), (512, Y*2), (778, Y*2),
                (256, Y*3), (512, Y*3), (778, Y*3)]
        ######################################
        # Black hole:
        if Global.Objective.Type == ObjectiveType.BlackHoleToss or 1:
            Locus = random.choice(Loci)
            Loci.remove(Locus)
            BlackHole = Sprite.AnimateSprite("BlackHole", Locus[0], Locus[1])
            self.ForegroundSprites.add(BlackHole)
            self.HazardSprites.add(BlackHole)
        ######################################
        # Walls:
        WallCount = random.choice((0, 0, 0, 0, 0, 1, 1, 1, 2, 2, 3))
        for Index in range(WallCount):
            Locus = random.choice(Loci)
            Loci.remove(Locus)
            Wall = Sprite.WallSprite(Locus[0] - 20, Locus[1] - 20, Locus[0] + 20, Locus[1] + 20)
            self.WallSprites.add(Wall)
            self.BackgroundSprites.add(Wall)
        ######################################
        # Space stations:
        # If it's a quest to kill snakes or gather items, build no spawners...
        if Global.Objective.Type != ObjectiveType.SnakePit:
            SSLoci = [(100, 60, 1, 0), (700, 60, 1, 0), (512, 660, 1, 0), (70, 400, 0, 1), (960, 400,0,1)]
            if Global.Objective.Type == ObjectiveType.KillType:
                TotalSpawners = random.choice((2, 3, 4))
            else:
                TotalSpawners = random.choice((1, 2))
            for Index in range(TotalSpawners):
                Locus = random.choice(SSLoci)
                SSLoci.remove(Locus)
                DXMult = random.choice((1, -1))
                DYMult = random.choice((1, -1))
                StationSprite = Alien.SpaceStationSprite(Locus[0], Locus[1], DXMult*Locus[2], DYMult*Locus[3])
                self.ForegroundSprites.add(StationSprite)
                self.MonsterSprites.add(StationSprite)
    def BuildSnakes(self):
        Loci = [(128, 32, 0), (512, 32, 1), (672, 32, 2),
                (128, 672, 0), (512, 672, 1), (672, 672, 2),]
        for Index in range(Global.Objective.Count):
            Locus = random.choice(Loci)
            SnakeHead = Snake.SnakeHead(*Locus)
            self.MonsterSprites.add(SnakeHead)
            self.ForegroundSprites.add(SnakeHead)
            self.SnakeSprites.add(SnakeHead)
    def MonkeyWithLawsOfPhysics(self, SectorNumber):
        "Establish the laws of physics and such for this sector"
        self.SectorNumber = SectorNumber
        random.seed(SectorNumber)
        self.ItemRate = random.choice((0.0, 0.02, 0.03, 0.05, 0.05, 0.05, 0.5))
        #self.ItemRate = 0.9 #%%%
        # Establish the laws of physics for this sector:
        if Global.Rules.Gravity == Global.GravityRule.Zero:
            self.HandleGravity = self.HandleGravityNone
        elif Global.Rules.Gravity == Global.GravityRule.Down:
            self.HandleGravity = self.HandleGravityDown
        elif Global.Rules.Gravity == Global.GravityRule.Up:
            self.HandleGravity = self.HandleGravityUp
        elif Global.Rules.Gravity == Global.GravityRule.Left:
            self.HandleGravity = self.HandleGravityLeft
        elif Global.Rules.Gravity == Global.GravityRule.Right:
            self.HandleGravity = self.HandleGravityRight
        # Create the player sprite:
        self.PlayerSprite = Player.PlayerSprite(400, 400)
        Global.PlayerSprite = self.PlayerSprite
        self.ForegroundSprites.add(self.PlayerSprite)
        self.BouncySprites.add(self.PlayerSprite)
        # Select alien behavior:
        self.SelectAlienBehavior(SectorNumber)
        # Build walls:
        self.BuildWallSprites(SectorNumber)
        if Global.Objective.Type == ObjectiveType.SnakePit:
            self.BuildSnakes()
        # Build status panel:
        self.BuildStatusPanel()
        # Build spawners and hazards:
        self.BuildSpawners()
        # Set key control:
        self.KeyHandlers[KeyCommand.Up] = self.PlayerSprite.Thrust
        self.KeyHandlers[KeyCommand.Left] = self.PlayerSprite.RotateLeft
        self.KeyHandlers[KeyCommand.Right] = self.PlayerSprite.RotateRight
        if Global.Rules.Down == Global.DownRule.Reverse:
            self.KeyHandlers[KeyCommand.Down] = self.PlayerSprite.Brake
            self.KeyRefractory[KeyCommand.Down] = 2
        elif Global.Rules.Down == Global.DownRule.ButtTorp:
            self.KeyHandlers[KeyCommand.Down] = self.HandleFireButtTorp
            self.KeyRefractory[KeyCommand.Down] = 10
        else:
            self.KeyHandlers[KeyCommand.Down] = self.HandleEtherBrakes
            self.KeyRefractory[KeyCommand.Down] = 20
        self.KeyHandlers[KeyCommand.Shot1] = self.HandleShootBasic
        self.ShieldsAvailable = 0
        self.StrafeAvailable = 0
        if Global.Rules.Button2 == Global.Button2Rule.Pickup:
            self.KeyHandlers[KeyCommand.Shot2] = self.HandleButton2Pickup
        elif Global.Rules.Button2 == Global.Button2Rule.Panic:
            self.KeyHandlers[KeyCommand.Shot2] = self.HandleButton2Panic
        elif Global.Rules.Button2 == Global.Button2Rule.Shield:
            self.KeyHandlers[KeyCommand.Shot2] = self.HandleButton2None
            self.ShieldsAvailable = 1
        elif Global.Rules.Button2 == Global.Button2Rule.Strafe:
            self.StrafeAvailable = 1
            self.KeyHandlers[KeyCommand.Shot2] = self.HandleButton2None
        else:
            self.KeyHandlers[KeyCommand.Shot2] = self.HandleButton2None
        self.RedrawSelf()
        self.PauseSprite = Sprite.TextSprite("Get ready...", 512, 300, Color.Green, 48)
        Global.CurrentSong = None
        #pygame.mixer.music.fadeout(400)
        pygame.mixer.music.pause()
        if random.random() > 0.8: 
            Sound.PlaySound(SoundNames.Obey)
        self.ForegroundSprites.add(self.PauseSprite)
        self.Countdown = 100
        self.Activity = Activity.StartLevelDelay
    def HandleButton2None(self):
        pass
    def HandleButton2Pickup(self):
        "Pickup/drop"
        if self.PlayerSprite.CarriedCritter:
            self.PlayerSprite.CarriedCritter.DX = self.PlayerSprite.XDir * 3
            self.PlayerSprite.CarriedCritter.DY = self.PlayerSprite.YDir * 3
            self.PlayerSprite.CarriedCritter.IsGrabbed = 0
            self.PlayerSprite.CarriedCritter.IsTossed = 1
            self.PlayerSprite.CarriedCritter = None
            Sound.PlaySound(SoundNames.PutDown)
        else:
            Bob = PointSprite(pygame.Rect(self.PlayerSprite.rect.left, self.PlayerSprite.rect.top,
                                          self.PlayerSprite.rect.width, self.PlayerSprite.rect.height))
            Sprites = pygame.sprite.spritecollide(Bob, self.MonsterSprites, 0)
            del Bob 
            for Sprite in Sprites:
                if not Sprite.IsHostile:
                    if Global.Objective.Type == ObjectiveType.Rescue:
                        Sound.PlaySound(SoundNames.PickUp)
                        Sprite.kill()
                        Global.Player.QuestProgress += 1
                        Global.Screen.QuestUpdate()                                            
                        return
                    else:
                        Sprite.IsGrabbed = 1
                        Sprite.GrabX = Sprite.rect.center[0] - self.PlayerSprite.X
                        Sprite.GrabY = Sprite.rect.center[1] - self.PlayerSprite.Y
                        Sound.PlaySound(SoundNames.PickUp)
                        self.PlayerSprite.CarriedCritter = Sprite
                        return
    def HandleButton2Panic(self):
        if Global.Player.Energy <= 100:
            Sound.PlaySound(SoundNames.Denied)
            return
        Global.Player.Energy -= 100
        Sound.PlaySound(SoundNames.Panic)
        Angle = 0
        while Angle < 2*math.pi:
            DX = math.cos(Angle)
            DY = math.sin(Angle)
            Shot = Sprite.ShotSprite(ImageNames.Panic, self.PlayerSprite.X + 2*DX,
                                    self.PlayerSprite.Y + 2*DY,
                                    DX*5, DY*5)
            Shot.Damage = 3
            self.AllShotSprites.add(Shot)
            self.ForegroundSprites.add(Shot)
            self.PlayerShotSprites.add(Shot)
            self.PanicSprites.add(Shot)
            Angle += 0.35
    def BuildStatusPanel(self):
        for StatusSprite in self.StatusPanelSprites.sprites():
            StatusSprite.kill()
        if not self.StatusPanelSprite:
            Image = pygame.Surface((Global.App.Width, Global.App.Height - self.Bottom))
            self.StatusPanelSprite = Sprite.Sprite(Image, 0, self.Bottom)
            self.BackgroundSprites.add(self.StatusPanelSprite)
        TheSprite = Sprite.TextSprite("Sector:", 100, self.Bottom + 10, Color.Grey, 24)
        self.BackgroundSprites.add(TheSprite)
        self.StatusPanelSprites.add(TheSprite)
        TheSprite = Sprite.TextSprite(str(self.SectorNumber), 120, self.Bottom + 35, Color.Grey, 24)
        self.BackgroundSprites.add(TheSprite)
        self.StatusPanelSprites.add(TheSprite)
        TheSprite = Sprite.TextSprite("Lives:", 200, self.Bottom + 10, Color.Grey, 24)
        self.BackgroundSprites.add(TheSprite)
        self.StatusPanelSprites.add(TheSprite)
        TheSprite = Sprite.TextSprite(str(Global.Player.Lives), 220, self.Bottom + 35, Color.Grey, 24)
        self.BackgroundSprites.add(TheSprite)
        self.StatusPanelSprites.add(TheSprite)
        ObjectiveString = GetObjectiveDescription(Global.Objective)
        TheSprite = Sprite.TextSprite(ObjectiveString, 400, self.Bottom + 10, Color.Grey, 24)
        self.BackgroundSprites.add(TheSprite)
        self.StatusPanelSprites.add(TheSprite)
        if not self.ObjectiveProgressSprite:
            self.ObjectiveProgressSprite = Sprite.TextSprite("", 420, self.Bottom + 35, Color.White, 24)
            self.ForegroundSprites.add(self.ObjectiveProgressSprite)
        ObjectiveProgress = "%s / %s"%(min(Global.Player.QuestProgress, Global.Objective.Count), Global.Objective.Count)
        self.ObjectiveProgressSprite.UpdateText(ObjectiveProgress)
        
        ObjectiveString = "Kill stuff"
        TheSprite = Sprite.TextSprite("Energy:", 650, self.Bottom + 10, Color.Grey, 24)
        self.BackgroundSprites.add(TheSprite)
        self.StatusPanelSprites.add(TheSprite)
        if not self.EnergySprite:
            self.EnergySprite = Sprite.EnergySprite(660, self.Bottom + 35)
            self.ForegroundSprites.add(self.EnergySprite)
        TheSprite = Sprite.TextSprite("Score:", 800, self.Bottom + 10, Color.Grey, 24)
        self.BackgroundSprites.add(TheSprite)
        self.StatusPanelSprites.add(TheSprite)
        if not self.ScoreSprite:
            self.ScoreSprite = Sprite.ScoreSprite(800, self.Bottom + 35)
            self.ForegroundSprites.add(self.ScoreSprite)
            
    def BuildWallSprites(self, SectorNumber):
        # Default walls:
        WallSprite = Sprite.WallSprite(0, 0, Global.App.Width, 10)
        self.WallSprites.add(WallSprite)
        self.BackgroundSprites.add(WallSprite)
        WallSprite = Sprite.WallSprite(0, self.Bottom - 10, Global.App.Width, self.Bottom)
        self.WallSprites.add(WallSprite)
        self.BackgroundSprites.add(WallSprite)
        WallSprite = Sprite.WallSprite(0, 0, 10, self.Bottom)
        self.WallSprites.add(WallSprite)
        self.BackgroundSprites.add(WallSprite)
        WallSprite = Sprite.WallSprite(Global.App.Width - 10, 0, Global.App.Width, self.Bottom)
        self.WallSprites.add(WallSprite)
        self.BackgroundSprites.add(WallSprite)
        # Other walls? %%%
    def SpawnIceCreamTruck(self):
        Sound.PlaySong("IceCreamTruck.xm")
        Truck = Alien.IceCreamSprite()
        self.ForegroundSprites.add(Truck)
    def HandleLoop(self):
        self.Tick = self.Tick + 1
        if self.Tick >= self.MaxTick:
            self.Tick = self.Tick % self.MaxTick
            if self.Activity == Activity.Play and random.random() > 0.9:
                self.SpawnIceCreamTruck()
        self.AnimationSprites.clear(self.Surface, self.BackgroundSurface)
        self.ForegroundSprites.clear(self.Surface, self.BackgroundSurface)
        self.Update()
        if self.ScreenClosed:
            return
        #if self.State == Activity.Play:
        self.HandleEvents()
        if self.ScreenClosed:
            return
        
        # Avoid juking chugging processor:
        if self.Activity == Activity.Pause:
            time.sleep(0.05)
        # Animation sprites are painted over foreground sprites.
        # Other than that, Z order is up to the magic pixies.
        DirtRects = self.ForegroundSprites.draw(self.Surface)
        MarkDirty(DirtRects)
        DirtRects = self.AnimationSprites.draw(self.Surface)
        MarkDirty(DirtRects)
    def HandleCollisions(self):
        #############################
        # Player bullets and critters:
        CollideDict = pygame.sprite.groupcollide(self.PlayerShotSprites, self.MonsterSprites, 0, 0)
        for (Shot, VictimList) in CollideDict.items():
            Victim = VictimList[0]
            DeadFlag = Victim.HandleShot(Shot)
            if DeadFlag:
                Victim.kill()
        if not self.PlayerSprite.Invincible:
            #############################
            # Monster bullets and player:            
            BulletSprites = pygame.sprite.spritecollide(self.PlayerSprite, self.MonsterShotSprites, 1)
            for BulletSprite in BulletSprites:
                self.PlayerSprite.HandleShot(BulletSprite)
                Global.Player.Invincible = 20
            #############################
            # Panic sprites:
            pygame.sprite.groupcollide(self.MonsterShotSprites, self.PanicSprites, 1, 0)
            #############################
            # Monsters and player:
            CollideSprites = pygame.sprite.spritecollide(self.PlayerSprite, self.MonsterSprites, 0)
            for CollideSprite in CollideSprites:
                if CollideSprite.IsHostile:
                    CollideSprite.CrashIntoPlayer()
                    Global.PlayerSprite.DX += CollideSprite.DX
                    Global.PlayerSprite.DX = min(10, max(-10, Global.PlayerSprite.DX))
                    Global.PlayerSprite.DY += CollideSprite.DY
                    Global.PlayerSprite.DY = min(10, max(-10, Global.PlayerSprite.DY))
                    Sound.PlaySound(SoundNames.PlayerHit)
                    if self.PlayerSprite.ShieldSprite:
                        Global.Player.Energy -= CollideSprite.Damage / 10
                    else:
                        Global.Player.Energy -= CollideSprite.Damage
                    Global.Player.Invincible = 20
                    self.CheckDeath()
            #############################
            # Player and items:  (Yes, items are affected by invincibility; otherwise you might
            # get poisoned and feel sad)  ("I'm sad now" - Hubert Farnsworth)
            CollideSprites = pygame.sprite.spritecollide(self.PlayerSprite, self.ItemSprites, 0)
            for CollideSprite in CollideSprites:
                CollideSprite.GetItem()
                    
        #############################
        # Shots and walls:
        pygame.sprite.groupcollide(self.AllShotSprites, self.WallSprites, 1, 0)
        pygame.sprite.groupcollide(self.ShrapnelSprites, self.WallSprites, 1, 0)
        #############################
        # Everyone else and walls:
        CollideDict = pygame.sprite.groupcollide(self.BouncySprites, self.WallSprites, 0, 0)
        for (BouncySprite, Walls) in CollideDict.items():
            for Wall in Walls:
                BouncySprite.BounceOffWall(Wall)
        #############################
        # Hazards and player/monsters/items:
        CollideDict = pygame.sprite.groupcollide(self.BouncySprites, self.HazardSprites, 0, 0)
        for (BouncySprite, Hazards) in CollideDict.items():
            # Assume all black holes.
            if isinstance(BouncySprite, Items.ItemSprite):
                Plummet = Sprite.PlummetSprite(BouncySprite)
                self.AnimationSprites.add(Plummet)
                BouncySprite.kill()
            elif BouncySprite == Global.PlayerSprite:
                Global.Player.Energy -= 5
                self.CheckDeath()
            else:
                if BouncySprite.IsTossed:
                    Sound.PlaySound(SoundNames.Plummet)
                    Plummet = Sprite.PlummetSprite(BouncySprite)
                    self.AnimationSprites.add(Plummet)                    
                    BouncySprite.kill()
                    Global.Player.QuestProgress += 1
                    Global.Screen.QuestUpdate()                    
            #BouncySprite.BounceOffWall(Wall)
        
    def QuestUpdate(self):
        #self.BuildStatusPanel()
        ObjectiveProgress = "%s / %s"%(min(Global.Player.QuestProgress, Global.Objective.Count), Global.Objective.Count)
        self.ObjectiveProgressSprite.UpdateText(ObjectiveProgress)
        self.RedrawSelf()
        if Global.Player.QuestProgress >= Global.Objective.Count:
            self.Activity = Activity.LevelComplete
            pygame.mixer.music.pause()
            self.Countdown = 100
            Sound.PlaySound(SoundNames.LevelComplete)
            self.PauseSprite = Sprite.TextSprite("Level Complete!", 512, 300, Color.Green, 48)
            self.ForegroundSprites.add(self.PauseSprite)
    def CheckDeath(self):
        if Global.Player.Energy > 0:
            return
        # DIE! DIE DIE DIE DIE DIE!
        Global.Player.Energy = 0
        pygame.mixer.music.pause()
        Sound.PlaySound(SoundNames.PlayerDie)
        self.AddShrapnel(self.PlayerSprite.X, self.PlayerSprite.Y, 35)
        if self.PlayerSprite.ShieldSprite:
            self.PlayerSprite.ShieldSprite.kill()
            self.PlayerSprite.ShieldSprite = None
        self.PlayerSprite.InvisoImage()
        self.PlayerSprite.Invincible = 50
        Global.Player.Lives -= 1        
        self.BuildStatusPanel()
        self.RedrawSelf()
        self.Activity = Activity.DeathDelay
        self.DeathDelayCounter = 100
        # Auto-toss [tm] - any passengers are thrown clear of the rotting hulk.  ROTTING HULK!
        if self.PlayerSprite.CarriedCritter:
            self.PlayerSprite.CarriedCritter.IsGrabbed = 0
            self.PlayerSprite.CarriedCritter.IsTossed = 1
            self.PlayerSprite.CarriedCritter.DX = self.PlayerSprite.XDir * 3
            self.PlayerSprite.CarriedCritter.DY = self.PlayerSprite.YDir * 3
            self.PlayerSprite.CarriedCritter = None
    def Update(self):
        if self.Activity == Activity.Play:
            # Conditions fade away.  The candle that burns twice as bright burns half
            # as long...
            for (Key, Time) in Global.Player.Conditions.items():
                Time = max(-1, Time - 1)
                if Key == Conditions.Poison and self.Tick%3 == 0:
                    Global.Player.Energy -= 1
                    self.CheckDeath()
                elif Key == Conditions.Regeneration and self.Tick%5 == 0:
                    Global.Player.Energy = min(1000, Global.Player.Energy + 1)
                if Time == 0:
                    del Global.Player.Conditions[Key]
                    self.PlayerSprite.UpdateSpeedLimit()
                    if Key == Conditions.Confusion:
                        # Back to normal!
                        self.KeyHandlers[KeyCommand.Left] = self.PlayerSprite.RotateLeft
                        self.KeyHandlers[KeyCommand.Right] = self.PlayerSprite.RotateRight
                else:
                    Global.Player.Conditions[Key] = Time
            for (Key, Time) in self.KeyDelays.items():
                self.KeyDelays[Key] = max(0, Time - 1)
            if self.PlayerSprite.ShieldSprite and self.Tick%10==0:
                Global.Player.Energy -= 1
                self.CheckDeath()
            self.ShotDelay = max(0, self.ShotDelay - 1)
            self.HandleHeldKeys()
            self.HandleCollisions()
            self.HandleGravity()
            for Sprite in self.ForegroundSprites.sprites():
                Sprite.Update()
            for Sprite in self.AnimationSprites.sprites():
                Sprite.Update()
        elif self.Activity == Activity.LevelComplete:
            self.Countdown -= 1
            for Sprite in self.ShrapnelSprites.sprites():
                Sprite.Update()            
            if self.Countdown <= 0:
                Global.Player.SectorNumber += 1 # Zoom!
                Global.Player.MaxSectorNumber = max(Global.Player.MaxSectorNumber, Global.Player.SectorNumber)
                self.DestroyScreen()
        elif self.Activity == Activity.StartLevelDelay:
            self.Countdown -= 1
            if self.Countdown <= 0:
                self.Activity = Activity.Play
                SongName = random.choice(("Ingame.xm", "Ingame2.xm"))
                Sound.PlaySong(SongName)
                pygame.mixer.music.set_endevent(EndSongEvent)
                if self.PauseSprite:
                    self.PauseSprite.kill()
                    self.PauseSprite = None
        elif self.Activity == Activity.DeathDelay:
            self.DeathDelayCounter -= 1
            for Sprite in self.ShrapnelSprites.sprites():
                Sprite.Update()
            if self.DeathDelayCounter == 50 and Global.Player.Lives > 0:
                self.PlayerSprite.SetImage()
                self.PlayerSprite.DX = 0
                self.PlayerSprite.DY = 0
                self.PlayerSprite.X = 512
                self.PlayerSprite.Y = 50
                self.PlayerSprite.rect.center = (self.PlayerSprite.X, self.PlayerSprite.Y)
            if self.DeathDelayCounter == 0:
                if Global.Player.Lives <= 0:
                    self.DestroyScreen()
                else:
                    Global.Player.Energy = 1000
                    self.PlayerSprite.SetImage()
                    Global.Player.Conditions = {}
                    self.Activity = Activity.Play
                    if self.PauseSprite:
                        self.PauseSprite.kill()
                        self.PauseSprite = None
                    pygame.mixer.music.unpause()
                
    def HandleEvents(self):
        for Event in pygame.event.get():
            self.HandleEvent(Event)
    def HandleHeldKeys(self):
        Keys = pygame.key.get_pressed()
        if (Keys[273] or Keys[264]) and not self.KeyDelays[KeyCommand.Up]:
            apply(self.KeyHandlers[KeyCommand.Up])
            self.KeyDelays[KeyCommand.Up] = self.KeyRefractory[KeyCommand.Up]
        if (Keys[274] or Keys[258]) and not self.KeyDelays[KeyCommand.Down]:
            apply(self.KeyHandlers[KeyCommand.Down])
            self.KeyDelays[KeyCommand.Down] = self.KeyRefractory[KeyCommand.Down]
        if (Keys[276] or Keys[260]) and not self.KeyDelays[KeyCommand.Left]:
            apply(self.KeyHandlers[KeyCommand.Left])
            self.KeyDelays[KeyCommand.Left] = self.KeyRefractory[KeyCommand.Left]
        if (Keys[275] or Keys[262]) and not self.KeyDelays[KeyCommand.Right]:
            apply(self.KeyHandlers[KeyCommand.Right])
            self.KeyDelays[KeyCommand.Right] = self.KeyRefractory[KeyCommand.Right]
        if (Keys[306] and not self.ShotDelay and not self.KeyDelays[KeyCommand.Shot1]):
            apply(self.KeyHandlers[KeyCommand.Shot1])
            # Handler sets delay.
        if (Keys[307] and not self.ShotDelay and not self.KeyDelays[KeyCommand.Shot1]):
            apply(self.KeyHandlers[KeyCommand.Shot1])
            # Handler sets delay.
        if self.ShieldsAvailable:
            if Keys[308] and not self.PlayerSprite.ShieldSprite:
                self.PlayerSprite.ShieldSprite = Player.ShieldSprite()
                self.ForegroundSprites.add(self.PlayerSprite.ShieldSprite)
                Sound.PlaySound(SoundNames.ShieldsUp)
            elif not Keys[308] and self.PlayerSprite.ShieldSprite:
                self.PlayerSprite.ShieldSprite.kill()
                self.PlayerSprite.ShieldSprite = None
    def HandleKey(self, Key):
##        if Key == ord("t"):
##            self.SpawnIceCreamTruck() #%%%
        if Key == 13:
            KeyMods = pygame.key.get_mods()
            if (KeyMods & (pygame.KMOD_RALT | pygame.KMOD_LALT)):
                Global.App.ToggleFullScreen()
        elif Key is pygame.K_ESCAPE:
            if self.Activity == Activity.Pause:
                Sound.PlaySound(SoundNames.Pause)
                self.PauseSprite.kill()
                self.PauseSprite2.kill()
                self.PauseSprite = None
                self.Activity = Activity.Play
                pygame.mixer.music.unpause()
            elif self.Activity == Activity.Play:
                Sound.PlaySound(SoundNames.Pause)
                self.Activity = Activity.Pause
                pygame.mixer.music.pause()
                self.PauseSprite = Sprite.TextSprite("Paused!", 512, 300, Color.Blue, 72)
                self.PauseSprite2 = Sprite.TextSprite("(Press q to quit)", 512, 375, Color.Blue, 24)
                self.ForegroundSprites.add(self.PauseSprite)
                self.ForegroundSprites.add(self.PauseSprite2)
        if self.Activity == Activity.Play:
            if Key == 306 and not self.ShotDelay:
                apply(self.KeyHandlers[KeyCommand.Shot1])
                self.KeyDelays[KeyCommand.Shot1] = self.KeyRefractory[KeyCommand.Shot1]
                self.ShotDelay = 8
            if Key == 308:
                apply(self.KeyHandlers[KeyCommand.Shot2])
##            Delay = self.KeyDelays.get(Key, 0)
##            if Delay:
##                return # No more right now!
##            Handler = self.KeyHandlers.get(Key, None)
##            if Handler:
##                apply(Handler)
        elif self.Activity == Activity.Pause:
            if Key == ord("q"):
                self.PauseSprite2.UpdateText("Really quit? (Y/N)")
            elif Key == ord("n"):
                self.PauseSprite2.UpdateText("(Press q to quit)")
            elif Key == ord("y"):
                if self.PauseSprite2.Text == "Really quit? (Y/N)":
                    Global.Player.Lives = 0 # signal end of play
                    self.DestroyScreen()
        elif self.Activity == Activity.StartLevelDelay:
            if Key == ord(" "):
                self.Countdown = 1
            
                
    def HandleEvent(self, Event):
        if Event.type is pygame.QUIT:
            Global.App.QuittingTime = 1
        elif Event.type == EndSongEvent:
            Sound.RepeatSong()
        elif Event.type is pygame.KEYDOWN:
            self.HandleKey(Event.key)
    def DestroyScreen(self):
        for Sprite in self.ForegroundSprites.sprites():
            Sprite.kill()
        for Sprite in self.BackgroundSprites.sprites():
            Sprite.kill()
        self.ScreenClosed = 1
    def HandleGravity(self): # SWAPPABLE
        pass # this function is reassigned
    def HandleGravityNone(self): #PLUGIN
        pass
    def HandleGravityDown(self): #PLUGIN
        self.PlayerSprite.DY = min(self.PlayerSprite.DY + 0.02, 10)
        for ShotSprite in self.PlayerShotSprites.sprites():
            ShotSprite.DY = min(ShotSprite.DY + 0.04, 10)
        for ItemSprite in self.ItemSprites.sprites():
            ItemSprite.DY = min(ItemSprite.DY + 0.04, 10)
    def HandleGravityUp(self): #PLUGIN
        self.PlayerSprite.DY = max(self.PlayerSprite.DY - 0.02, -10)
        for ShotSprite in self.PlayerShotSprites.sprites():
            ShotSprite.DY = max(ShotSprite.DY - 0.04, -10)
        for ItemSprite in self.ItemSprites.sprites():
            ItemSprite.DY = max(ItemSprite.DY - 0.04, -10)
    def HandleGravityRight(self): #PLUGIN
        self.PlayerSprite.DX = min(self.PlayerSprite.DX + 0.02, 10)
        for ShotSprite in self.PlayerShotSprites.sprites():
            ShotSprite.DX = min(ShotSprite.DX + 0.04, 10)
        for ItemSprite in self.ItemSprites.sprites():
            ItemSprite.DX = min(ItemSprite.DX + 0.04, 10)
    def HandleGravityLeft(self): #PLUGIN
        self.PlayerSprite.DX = max(self.PlayerSprite.DX - 0.02, -10)
        for ShotSprite in self.PlayerShotSprites.sprites():
            ShotSprite.DX = max(ShotSprite.DX - 0.04, -10)
        for ItemSprite in self.ItemSprites.sprites():
            ItemSprite.DX = max(ItemSprite.DX - 0.04, -10)
    def HandleEtherBrakes(self): #PLUGIN
        self.KeyDelays[KeyCommand.Down] = self.KeyRefractory[KeyCommand.Down]
        self.PlayerSprite.DX = 0
        self.PlayerSprite.DY = 0
        Sound.PlaySound(SoundNames.Brakes)
    def HandleFireButtTorp(self): #PLUGIN
        self.KeyDelays[KeyCommand.Down] = self.KeyRefractory[KeyCommand.Down]
        Shot = Sprite.ShotSprite(ImageNames.ButtTorp, self.PlayerSprite.X - 5*self.PlayerSprite.XDir,
                                    self.PlayerSprite.Y - 5*self.PlayerSprite.YDir,
                                    self.PlayerSprite.XDir * -5, self.PlayerSprite.YDir * -5)
        self.AllShotSprites.add(Shot)
        self.ForegroundSprites.add(Shot)
        self.PlayerShotSprites.add(Shot)
        Sound.PlaySound(SoundNames.ButtTorp)
    def HandleShootBasic(self): # SWAPPABLE
        self.KeyDelays[KeyCommand.Shot1] = self.KeyRefractory[KeyCommand.Shot1]
        if Global.Player.Conditions.get(Conditions.Shotgun):
            TheRange = random.choice((range(-2, 3), range(-1, 2)))
            for DirDelta in TheRange:
                Dir = self.PlayerSprite.Dir + DirDelta
                if (Dir < 0):
                    Dir += len(Player.XDir)
                elif Dir >= len(Player.XDir):
                    Dir -= len(Player.XDir)
                XDir = Player.XDir[Dir]
                YDir = Player.YDir[Dir]
                Shot = Sprite.ShotSprite(ImageNames.ShotGun, self.PlayerSprite.X + 5*self.PlayerSprite.XDir,
                                            self.PlayerSprite.Y + 5*self.PlayerSprite.YDir,
                                            XDir * 6, YDir * 6)
                self.AllShotSprites.add(Shot)
                self.ForegroundSprites.add(Shot)
                self.PlayerShotSprites.add(Shot)
            Sound.PlaySound(SoundNames.ShotGun)    
            self.ShotDelay = 14
        elif Global.Player.Conditions.get(Conditions.Phaser):
            for Times in range(1, 5):
                Shot = Sprite.ShotSprite(ImageNames.Phaser, self.PlayerSprite.X + 3*Times*self.PlayerSprite.XDir,
                                            self.PlayerSprite.Y + 3*Times*self.PlayerSprite.YDir,
                                            self.PlayerSprite.XDir * 8, self.PlayerSprite.YDir * 8)
                self.AllShotSprites.add(Shot)
                self.ForegroundSprites.add(Shot)
                self.PlayerShotSprites.add(Shot)
            Sound.PlaySound(SoundNames.Phaser)
            self.ShotDelay = 20
        else:
            Shot = Sprite.ShotSprite(ImageNames.Shot, self.PlayerSprite.X + 5*self.PlayerSprite.XDir,
                                        self.PlayerSprite.Y + 5*self.PlayerSprite.YDir,
                                        self.PlayerSprite.XDir * 5, self.PlayerSprite.YDir * 5)
            self.AllShotSprites.add(Shot)
            self.ForegroundSprites.add(Shot)
            self.PlayerShotSprites.add(Shot)
            Sound.PlaySound(SoundNames.Shot)
            self.ShotDelay = 8
            
        