"""
STAR SNAKE!  (badger badger badger badger)
"""
from Utils import *
from Constants import *
import Sprite
import Global
import math
import Sound
import Items

SnakeDirs = [(1, 0), (0, 1), (-1, 0), (0, -1)]
OppositeDirs = {0:2, 1:3, 2:0, 3:1}
TurnDirs = {0:(1,3), 1:(0,2), 2:(1,3), 3:(0,2)}

class SnakeHead(Sprite.Sprite):
    StepSize = 16
    MaxCountdown = 20
    IsHostile = 1
    DX = 0
    DY = 0
    Damage = 100
    def __init__(self, X, Y, Dir):
        self.X = X
        self.Y = Y
        self.Dir = Dir
        self.HP = 20
        self.Countdown = self.MaxCountdown
        self.Length = 10
        self.TailSprites = []
        Sprite.Sprite.__init__(self, ImageNames.SnakeHead, X, Y)
    def Update(self):
        self.Countdown -= 1
        if self.Countdown > 0:
            return
        self.Countdown = self.MaxCountdown
        # Kill a tail sprite, probably:
        if len(self.TailSprites) >= self.Length:
            self.TailSprites[0].kill()
            del self.TailSprites[0]
        Dirs = list(TurnDirs[self.Dir])
        Dirs.append(self.Dir)
        Dirs.reverse()
        if random.random() > 0.94: 
            random.shuffle(Dirs)
        LegalMoveFlag = 0
        for TryDir in Dirs:
            NextX = self.X + SnakeDirs[TryDir][0]*32
            NextY = self.Y + SnakeDirs[TryDir][1]*32
            # Check if it's a legal move:
            Bob = PointSprite(pygame.Rect(NextX-16, NextY-16, 32, 32))
            CollideSprite = pygame.sprite.spritecollideany(Bob, Global.Screen.WallSprites)
            CollideSprite2 = pygame.sprite.spritecollideany(Bob, Global.Screen.SnakeSprites)
            del Bob
            if not CollideSprite and not CollideSprite2:
                LegalMoveFlag = 1
                break
        if not LegalMoveFlag:
            self.Die()
            return
        self.Dir = TryDir
        Segment = SnakeSegment(self)
        Global.Screen.MonsterSprites.add(Segment)
        Global.Screen.SnakeSprites.add(Segment)
        Global.Screen.ForegroundSprites.add(Segment)
        self.TailSprites.append(Segment)
        self.X = NextX
        self.Y = NextY
        #print self.X, self.Y, self.Dir
        self.rect.center = (self.X, self.Y)
    def Die(self):
        self.HP = 0
        Sound.PlaySound(SoundNames.SnakeDeath)
        Global.Screen.AddShrapnel(self.X, self.Y, 10)
        for TailSprite in self.TailSprites:
            TailSprite.Die()
        self.kill()
        Global.Player.QuestProgress += 1
        Global.Screen.QuestUpdate()        
    def HandleShot(self, ShotSprite):
        self.HP -= ShotSprite.Damage
        ShotSprite.kill()
        if self.HP <= 0:
            self.Die()
        else:
            Sound.PlaySound(SoundNames.MonsterHit)
    def CrashIntoPlayer(self):
        Global.PlayerSprite.BounceOffWall(self)
class SnakeSegment(Sprite.Sprite):
    IsHostile = 1
    DX = 0
    DY = 0
    Damage = 100
    def __init__(self, Head):
        self.X = Head.X
        self.Y = Head.Y
        self.Head = Head
        Sprite.Sprite.__init__(self, ImageNames.SnakeTail, self.X, self.Y)
    def Update(self):
        pass
    def Die(self):
        # Drop up to 15 capsules on screen at once.
        if random.random() <= Global.Screen.ItemRate and len(Global.Screen.ItemSprites.sprites()) < 15:
            Item = Items.ItemSprite(self.X, self.Y) 
            Global.Screen.ForegroundSprites.add(Item)
            Global.Screen.ItemSprites.add(Item)
            Global.Screen.BouncySprites.add(Item)
        Global.Screen.AddShrapnel(self.X, self.Y, 2)
        self.kill()
    def HandleShot(self, Shot):
        self.Head.Length = min(self.Head.Length + 1, 40) 
        Sound.PlaySound(SoundNames.ShotSnakeTail)
        Shot.kill()
    def CrashIntoPlayer(self):
        Global.PlayerSprite.BounceOffWall(self)
        