import pygame
import Global
import os
from Constants import *
class Sprite(pygame.sprite.Sprite):
    def __init__(self, Image, X, Y):
        if type(Image)==type(""):
            Image = Global.Images[Image][0]
        self.image = Image
        self.rect = Image.get_rect()
        self.image.set_colorkey(Color.Black)
        self.rect.center = (X,Y)
        pygame.sprite.Sprite.__init__(self)
    def ReplaceImage(self, NewImage):
        OldCenter=  self.rect.center
        self.image = NewImage
        self.rect = self.image.get_rect()
        self.rect.center = OldCenter
    def Update(self):
        pass
    def kill(self):
        pygame.sprite.Sprite.kill(self)
        self.rect = None
        self.image = None


class Bouncer:
    def WallReact(self, Wall):
        # This is called to do something INTERESTING when we hit a wall.
        pass
    def BounceOffWall(self, Wall):
        (X, Y) = self.rect.center
        if Wall.rect.left < X < Wall.rect.right:
            if self.rect.top < Wall.rect.top:
                self.DY = -abs(self.DY) * 0.5
                self.rect.bottom = min(Wall.rect.top, self.rect.bottom)
                self.Y = self.rect.center[1] # Prevent wallwalking!
            else:
                self.DY = abs(self.DY) * 0.5
                self.rect.top = max(Wall.rect.bottom, self.rect.top)
                self.Y = self.rect.center[1]
        elif Wall.rect.top < Y < Wall.rect.bottom:
            if self.rect.left < Wall.rect.left:
                self.DX = -abs(self.DX) * 0.5
                self.rect.right = min(Wall.rect.left, self.rect.right)
                self.X = self.rect.center[0]
            else:
                self.DX = abs(self.DX) * 0.5
                self.rect.left = max(Wall.rect.right, self.rect.left)
                self.X = self.rect.center[0]
        else:
            # Diagonal approach, diagonal bounce:
            if self.rect.top < Wall.rect.top:
                self.DY = -abs(self.DY) * 0.5
                self.rect.bottom = min(Wall.rect.top, self.rect.bottom)
                self.Y = self.rect.center[1]
            else:
                self.DY = abs(self.DY) * 0.5
                self.rect.top = max(Wall.rect.bottom, self.rect.top)
                self.Y = self.rect.center[1]
            if self.rect.left < Wall.rect.left:
                self.DX = -abs(self.DX) * 0.5
                self.rect.right = min(Wall.rect.left, self.rect.right)
                self.X = self.rect.center[0]
            else:
                self.DX = abs(self.DX) * 0.5
                self.rect.left = max(Wall.rect.right, self.rect.left)
                self.X = self.rect.center[0]
        self.WallReact(Wall)
    

# Extend Sprite class with some slightly useful stuff:
class AnimateSprite(Sprite):
    def __init__(self, ImageName, X, Y, TicksPerImage = 10):
        self.Images = Global.Images[ImageName]
        self.TicksPerImage = TicksPerImage
        self.ImageNumber = 0
        self.Tick = 0
        Sprite.__init__(self, self.Images[0], X, Y)
    def Update(self):
        self.Tick += 1
        if self.Tick > self.TicksPerImage:
            self.ImageNumber = (self.ImageNumber + 1)%len(self.Images)
            self.ReplaceImage(self.Images[self.ImageNumber])
            self.Tick = 0

class PlummetSprite(AnimateSprite):
    def __init__(self, Victim):
        (X, Y) = Victim.rect.center
        AnimateSprite.__init__(self, ImageNames.Plummet, X, Y)
    def Update(self):
        self.Tick += 1
        if self.Tick > self.TicksPerImage:
            self.ImageNumber = (self.ImageNumber + 1)
            if self.ImageNumber >= len(self.Images):
                self.kill()
                return
            self.ReplaceImage(self.Images[self.ImageNumber])
            self.Tick = 0
        

class ShotSprite(AnimateSprite):
    def __init__(self, Name, X, Y, DX, DY):
        AnimateSprite.__init__(self, Name, X, Y, 3)
        self.DX = DX
        self.DY = DY
        self.X = X
        self.Y = Y
        self.Age = 0
        self.Damage = 1
    def Update(self):
        self.X += self.DX
        self.Y += self.DY
        self.rect.center = (self.X, self.Y)
        # Kill the shot after a certain number of ticks, so that bouncy shots don't inherit the earth:
        self.Age += 1
        if self.Age > 200:
            self.kill()
        else:
            AnimateSprite.Update(self)

class ThrustSprite(AnimateSprite):
    def __init__(self, X, Y, DX, DY):
        self.X = X
        self.Y = Y
        self.DX = DX + random.randrange(-10, 10)/20.0
        self.DY = DY + random.randrange(-10, 10)/20.0
        self.Age = 0
        AnimateSprite.__init__(self, ImageNames.Exhaust, X, Y, 1)
    def Update(self):
        self.X += self.DX
        self.Y += self.DY
        self.rect.center = (self.X, self.Y)
        self.Age += 1
        if self.Age > 20:
            self.kill()
        else:
            AnimateSprite.Update(self)

class ShrapnelSprite(AnimateSprite):
    def __init__(self, X, Y):
        self.X = X
        self.Y = Y
        self.DX = random.choice((1, -1)) * random.randrange(20, 50)/20.0
        self.DY = random.choice((1, -1)) * random.randrange(20, 50)/20.0
        self.Age = random.randrange(20)
        AnimateSprite.__init__(self, ImageNames.Shrapnel, X, Y, 1)
    def Update(self):
        self.X += self.DX
        self.Y += self.DY
        self.rect.center = (self.X, self.Y)
        self.Age += 1
        if self.Age > 40:
            self.kill()
        else:
            AnimateSprite.Update(self)

class WallSprite(Sprite, Bouncer):
    def __init__(self, Left, Top, Right, Bottom, Texture = 0):
        self.image = pygame.Surface((Right-Left, Bottom-Top))
        self.image.fill(Color.White)
        self.rect = self.image.get_rect()
        self.rect.top = Top
        self.rect.left = Left
        Sprite.__init__(self, self.image, (Left+Right)/2, (Top+Bottom)/2)

def GetFont(Name, Size):
    if not Name:
        Name = "bluebold"
    FontName = os.path.join("Fonts", "%s.ttf"%Name)
    TheFont = pygame.font.Font(FontName, Size)
    return TheFont


class TextSprite(Sprite):
    def __init__(self, Text, X, Y, TextColor = Color.White, FontSize = 16, FontName = None):
        self.Text = Text
        self.TextColor = TextColor
        self.Font = GetFont(FontName, FontSize) 
        Image = self.Font.render(str(self.Text), 1, self.TextColor)
        Sprite.__init__(self,Image, X, Y)
    def UpdateText(self, NewText):
        self.Text = NewText
        OldCenter = self.rect.center
        self.image = self.Font.render(str(self.Text), 1, self.TextColor)
        self.rect = self.image.get_rect()
        self.rect.center = OldCenter

class ButtonSprite(Sprite):
    def __init__(self, Text, X, Y):
        self.Text = Text
        self.Font = GetFont(None, 24) 
        FontImage = self.Font.render(str(self.Text), 1, Color.Green)
        Width = FontImage.get_rect().width + 6
        Height = FontImage.get_rect().height + 4
        self.image = pygame.Surface((Width, Height))
        self.image.fill(Color.DarkGrey)
        pygame.draw.rect(self.image, Color.White, (0, 0, Width, Height), 1)
        
        self.image.blit(FontImage, (2, 1))
        Sprite.__init__(self, self.image, X, Y)
    
class EnergySprite(TextSprite):
    def __init__(self, X, Y):
        TextSprite.__init__(self, str(Global.Player.Energy), X, Y, Color.Green, 24)
    def Update(self):
        if Global.Player.Energy > 800:
            self.TextColor = Color.Green
        elif Global.Player.Energy > 300:
            self.TextColor = Color.Yellow
        else:
            self.TextColor = Color.Red
        self.UpdateText(str(Global.Player.Energy))

class ScoreSprite(TextSprite):
    def __init__(self, X, Y):
        TextSprite.__init__(self, str(Global.Player.Score), X, Y, Color.Green, 24)
    def Update(self):
        self.UpdateText(str(Global.Player.Score))
        