"""
Trouble Shooter - an entry for the October 2004 Ludum Dare 48-hour compo
Theme: "Random"
They want random?  I'll give them random!  A game where the laws of physics
change with every level, at the whim of SPACE POPE.
"""
import os
import sys
# Be sure we're running from the app directory:
if len(sys.argv)>0:
    Dir = os.path.split(sys.argv[0])[0]
    if Dir:
        os.chdir(Dir)
from Constants import *
from Utils import *
import pygame
import Images
import Sound
import Global
import Screen
import Alien
import LevelScreen
Global.FilthyFilthyRectangles = []

class TroubleApp:
    def __init__(self):
        self.QuittingTime = 0
        self.CurrentScreen = None
    def Initialize(self):
        pygame.init()
        pygame.mixer.init()
        Global.ResetItemEffects()
        Global.NullImage = pygame.Surface((1, 1))
        Global.NullImage.fill(Color.Black)
        Global.NullImage.set_colorkey(Color.Black)
        self.Width = 1024
        self.Height = 768
        # Sneaky trick: Go fullscreen immediately if we're an .exe
        if sys.argv[0][-3:] == ".py":
            self.Surface = pygame.display.set_mode((self.Width, self.Height), pygame.FULLSCREEN) 
            #self.Surface = pygame.display.set_mode((self.Width, self.Height)) #, pygame.FULLSCREEN) 
        else:
            self.Surface = pygame.display.set_mode((self.Width, self.Height), pygame.FULLSCREEN) 
        self.Surface.fill(Color.Black)
        self.BackgroundSurface = pygame.Surface((self.Width, self.Height))
        pygame.display.set_caption("Trouble Shooter")
        pygame.display.flip()
        self.Clock = pygame.time.Clock()
        Images.PrepImages()
        Sound.PrepSounds()
    def EnterSector(self, SectorNumber):
        self.GetObjective(SectorNumber)
        self.CurrentScreen = LevelScreen.LevelIntroScreen(self)
        self.CurrentScreen.Render()
        Global.Screen = self.CurrentScreen
        pygame.mouse.set_visible(1)
    def InitPlayScreen(self):
        self.CurrentScreen = Screen.SpaceScreen(self)
        Global.Screen = self.CurrentScreen
        self.CurrentScreen.MonkeyWithLawsOfPhysics(Global.Player.SectorNumber)
        self.CurrentScreen.BuildStatusPanel()
        pygame.mouse.set_visible(0)
    def StartGame(self, SectorNumber):
        Global.Player.SectorNumber = SectorNumber
        self.EnterSector(SectorNumber)
    def GetObjective(self, SectorNumber):
        random.seed(SectorNumber)
        Object = ObjectiveClass()
        Object.Type = random.choice((ObjectiveType.KillType, ObjectiveType.KillType, ObjectiveType.KillType,
                ObjectiveType.BlackHoleToss, ObjectiveType.BlackHoleToss,
                ObjectiveType.Rescue, ObjectiveType.SnakePit, ))
        if Object.Type in (ObjectiveType.KillType, ObjectiveType.Rescue, ObjectiveType.BlackHoleToss):
            Object.Class = random.choice((Alien.GreenAlienSprite, Alien.BlueAlienSprite, Alien.RedAlienSprite))
        else:
            Object.Class = None
        if Object.Type  == ObjectiveType.KillType:
            Object.Count = random.randrange(5, 20)
        elif Object.Type  == ObjectiveType.Rescue:
            Object.Count = random.randrange(5, 20)
        elif Object.Type  == ObjectiveType.BlackHoleToss:
            Object.Count = random.randrange(3, 10)
        elif Object.Type == ObjectiveType.SnakePit:
            Object.Count = random.randrange(3, 5)
        Global.Objective = Object
        Rules = Global.RuleClass()
        if Object.Type in (ObjectiveType.BlackHoleToss, ObjectiveType.Rescue):
            Rules.Button2 = Global.Button2Rule.Pickup
        else:
            Rules.Button2 = random.choice((Global.Button2Rule.Shield,
                                           Global.Button2Rule.Strafe,
                                           Global.Button2Rule.Panic))
        Rules.Gravity = random.choice((Global.GravityRule.Zero, Global.GravityRule.Zero, Global.GravityRule.Zero,
                                      Global.GravityRule.Zero, Global.GravityRule.Zero, Global.GravityRule.Zero,
                                      Global.GravityRule.Down, Global.GravityRule.Down, Global.GravityRule.Down,
                                      Global.GravityRule.Up,Global.GravityRule.Left,Global.GravityRule.Right,))
        Rules.Down = random.choice((Global.DownRule.ButtTorp, Global.DownRule.EtherBrakes,
                                        Global.DownRule.Reverse))
        Global.Rules = Rules
        Global.Player.QuestProgress = 0
        return Object
    def PlayGame(self):
        pygame.display.flip()
        while not self.QuittingTime:
            self.Clock.tick(60)
            pygame.event.pump()
            self.CurrentScreen.HandleLoop()
            pygame.display.update(Global.FilthyFilthyRectangles)
            Global.FilthyFilthyRectangles = []
            # If they alt-tabbed away, be sensible hobbits:
            while not pygame.display.get_active():
                pygame.time.wait(100)
                pygame.event.pump()
            if self.CurrentScreen.ScreenClosed:
                if Global.Player.Lives <= 0:
                    Global.ResetItemEffects()
                    Global.Player.Reset()
                if not self.QuittingTime:
                    if isinstance(self.CurrentScreen, Screen.SpaceScreen):
                        # Move to level title:
                        self.EnterSector(Global.Player.SectorNumber)
                    else:
                        self.InitPlayScreen()
        print "Leaving so soon?  You deserve %.2f lashes!"%self.Clock.get_fps()
        return
    def ToggleFullScreen(self):
        if (self.Surface.get_flags() & pygame.FULLSCREEN):
            Flags = 0
        else:
            Flags = pygame.FULLSCREEN
        CapturedSurface = pygame.Surface(self.Surface.get_size())
        CapturedSurface.blit(self.Surface, (0,0))
        ClipCapture = self.Surface.get_clip()
        self.Surface = pygame.display.set_mode((self.Width, self.Height), Flags)
        self.Surface.blit(CapturedSurface, (0,0))
        self.Surface.set_clip(ClipCapture)
        MarkDirty((self.Surface.get_rect(),))

def Main():
    Global.App = TroubleApp()
    Global.App.Initialize()
    # psyco goes here: %%%%
    if len(sys.argv)>1:
        SectorNumber= int(sys.argv[1])
    else:
        SectorNumber = 1
    Global.App.StartGame(SectorNumber)
    Global.App.PlayGame()
    
if __name__ == "__main__":
    Main()